/*
** Module   : Color.java
** Abstract : A color and attribute container.
**
** Copyright (c) 2005-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SIY 20050000  @22845  Created initial version
** 002 SIY 20110112          Changes related to new color handling scheme.
** 003 HC  20140119          Implemented runtime support for the COLOR 
**                           protermcap-attribute.
*/

package com.goldencode.p2j.ui;

import java.io.*;

/**
 * This represent single logical color as it is seend by the application code.
 * Interpretation of this color and mapping to actual terminal capabilities is
 * up to the low level driver.
 *
 * @author  SIY
 * @version 1.0.0
 */
public class Color
implements Externalizable
{
   /** Default color. */
   public static final Color DEFAULT = new Color();

   /** Color constant for NORMAL color. */
   public static final Color NORMAL   = DEFAULT;

   /** Color constant for INPUT color. */
   public static final Color INPUT    = DEFAULT.withUnderline();

   /** Color constant for MESSAGES color. */
   public static final Color MESSAGES = DEFAULT.withReverse();

   /** Color constant for BOLD color. */
   public static final Color BOLD     = DEFAULT.withBold();

   /** Color constant for BLINK color. */
   public static final Color BLINK    = DEFAULT.withBlink();

   /** Terminal color. */
   private int color;

   /** Terminal attribute. */
   private int attribute;
   
   /**
    * Construct an instance using specified color and attribute.
    * To construct the instance from a foreground and background
    * color components use 
    * {@link Color.#toColorValue(ColorPalette, ColorPalette)}.
    *
    * @param   attribute
    *          Terminal attribute.
    * @param   color
    *          Terminal color.
    */
   public Color(int attribute, int color)
   {
      this.color = color;
      this.attribute = attribute;
   }

   /**
    * Default constructor.
    */
   public Color()
   {
      color = 0;
      attribute = ColorAttribute.NORMAL.getValue();
   }

   /**
    * Create a copy of existing color.
    *
    * @param   other
    *          Source color.
    */
   private Color(Color other)
   {
      color = other.color;
      attribute = other.attribute;
   }

   /**
    * Get attribute.
    *
    * @return  attribute.
    */
   public int attribute()
   {
      return attribute;
   }

   /**
    * Get color.
    *
    * @return  color.
    */
   public int color()
   {
      return color;
   }

   /**
    * @see java.lang.Object#equals(java.lang.Object)
    */
   @Override
   public boolean equals(Object obj)
   {
      if (this == obj)
         return true;
      if (obj == null || getClass() != obj.getClass())
         return false;

      Color other = (Color) obj;

      return (attribute == other.attribute) && (color == other.color);
   }

   /**
    * Check if BLINK color property is active.
    *
    * @return <code>true</code> if property is active.
    */
   public boolean hasBlink()
   {
      return has(ColorAttribute.BLINK.getValue());
   }

   /**
    * Check if BOLD color property is active.
    *
    * @return <code>true</code> if property is active.
    */
   public boolean hasBold()
   {
      return has(ColorAttribute.BOLD.getValue());
   }

   /**
    * Check if REVERSE color property is active.
    *
    * @return <code>true</code> if property is active.
    */
   public boolean hasReverse()
   {
      return has(ColorAttribute.REVERSE.getValue());
   }

   /**
    * Check if UNDERLINE color property is active.
    *
    * @return <code>true</code> if property is active.
    */
   public boolean hasUnderline()
   {
      return has(ColorAttribute.UNDERLINE.getValue());
   }

   /**
    * Create color identical to this one but with active BLINK color property.
    *
    * @return new color.
    */
   public Color withBlink()
   {
      return copy().set(ColorAttribute.BLINK.getValue());
   }

   /**
    * Create color identical to this one but with active BOLD color property.
    *
    * @return new color.
    */
   public Color withBold()
   {
      return copy().set(ColorAttribute.BOLD.getValue());
   }

   /**
    * Create color identical to this one but with inactive BLINK color property.
    *
    * @return new color.
    */
   public Color withoutBlink()
   {
      return copy().reset(ColorAttribute.BLINK.getValue());
   }

   /**
    * Create color identical to this one but with inactive BOLD color property.
    *
    * @return new color.
    */
   public Color withoutBold()
   {
      return copy().reset(ColorAttribute.BOLD.getValue());
   }

   /**
    * Create color identical to this one but with inactive REVERSE color property.
    *
    * @return new color.
    */
   public Color withoutReverse()
   {
      return copy().reset(ColorAttribute.REVERSE.getValue());
   }

   /**
    * Create color identical to this one but with inactive UNDERLINE color
    * property.
    *
    * @return new color.
    */
   public Color withoutUnderline()
   {
      return copy().reset(ColorAttribute.UNDERLINE.getValue());
   }

   /**
    * Create color identical to this one but with active REVERSE color property.
    *
    * @return new color.
    */
   public Color withReverse()
   {
      return copy().set(ColorAttribute.REVERSE.getValue());
   }

   /**
    * Create color identical to this one but with active UNDERLINE color
    * property.
    *
    * @return new color.
    */
   public Color withUnderline()
   {
      return copy().set(ColorAttribute.UNDERLINE.getValue());
   }

   /**
    * @see java.lang.Object#hashCode()
    */
   @Override
   public int hashCode()
   {
      return 31 * (31 + attribute) + color;
   }

   /**
    * @see java.io.Externalizable#readExternal(java.io.ObjectInput)
    */
   @Override
   public void readExternal(ObjectInput in) throws IOException
   {
      color = in.readInt();
      attribute = in.readInt();
   }

   /**
    * @see java.io.Externalizable#writeExternal(java.io.ObjectOutput)
    */
   @Override
   public void writeExternal(ObjectOutput out) throws IOException
   {
      out.writeInt(color);
      out.writeInt(attribute);
   }

   /**
    * Returns the foreground component of the color value.
    * 
    * @param   color
    *          The {@link Color} instance.
    *          
    * @return
    *          The instance of {@link ColorPalette}.
    */
   public static ColorPalette toForeground(Color color)
   {
      int v = (color.color >> 16) - 1;
      return v < 0 || v >= ColorPalette.values().length ? null : ColorPalette.values()[v];
   }
   
   /**
    * Returns the background component of the color value.
    * 
    * @param   color
    *          The {@link Color} instance.
    *          
    * @return
    *          The instance of {@link ColorPalette}.
    */
   public static ColorPalette toBackground(Color color)
   {
      int v = (0x0000FFFF & color.color) - 1;
      return v < 0 || v >= ColorPalette.values().length ? null : ColorPalette.values()[v];
   }
   
   /**
    * Creates the color value from the foreground and the background
    * color components. The result integer is suitable to be used
    * in the constructor, @see Color#Color(int, int).  
    * 
    * @param   foreground
    *          The foreground color component.
    *          
    * @param   background
    *          The background color component.
    *          
    * @return
    *          The color integer value.
    */
   public static int toColorValue(ColorPalette foreground, ColorPalette background)
   {
      return ((foreground.ordinal() + 1) << 16) | (background.ordinal() + 1);
   }

   /**
    * Create copy of current color.
    *
    * @return  copy of current color.
    */
   private Color copy() {
      return new Color(this);
   }

   /**
    * Check bit in the color properties.
    *
    * @param   attr
    *          Property to check.
    *
    * @return  <code>true</code> if appropriate bit is set.
    */
   private boolean has(int attr) {
      return (attribute & attr) != 0;
   }

   /**
    * Reset specified bit in the color properties.
    *
    * @param   attr
    *          Property to reset.
    *
    * @return  this for call chaining.
    */
   private Color reset(int attr) {
      attribute &= ~attr;
      return this;
   }

   /**
    * Set specified bit in the color properties.
    *
    * @param   attr
    *          Property to set.
    *
    * @return  this for call chaining.
    */
   private Color set(int attr) {
      attribute |= attr;
      return this;
   }
}
