/*
** Module   : terminal.c
** Abstract : Generic implementation for character mode-based console functions.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 EVL 20130723 Created initial version to implement text console terminal JNI.
** 002 EVL 20130729 Creating native module for Windows, moving platform independent code here.
** 003 HC  20140119 Implemented runtime support for the COLOR protermcap-attribute.
*/

#include "terminal.h"

#include "com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper.h"

// These globals will be initialized in initConsole() function
int base_rows = 0;
int base_cols = 0;

// Externals definitions
extern jint attribute_result[ATTRIBUTE_MAX];
extern jint acs_result[ACS_MAX];

// Local function calls
/**
 * Throws the exception on the java side for given class name.
 *
 * @param   env        pointer to a structure that provides access to JNI environment.
 *
 * @param   name       the name of the target class for exception.
 * 
 * @param   msg        the message content of the exception.
 */
void throwException(JNIEnv* env, char* name, char* msg)
{
   jclass jcls = (*env)->FindClass(env, name);
   
   if (jcls != NULL)
   {
      (*env)->ThrowNew(env, jcls, msg);
   }
}

// The JNI calls
/** 
 * This function outputs the screenful of data at once using an array of codes where every
 * character position is either a skip indicator (-1) or a fully prepared code with character,
 * attribute and color included. The final step is the cursor positioning.
 *
 * @param   env      pointer to a structure that provides access to JNI environment.
 *
 * @param   jo       reference to the Java object who called this function.
 *
 * @param   chars    the array of characters to output. The array is supposed to provide an item
 *                   per every screen position. This item is either -1 if this position is masked
 *                   or a code which goes directly to ncurses.
 *
 * @param   attrs    the array of attributes to output. The array is supposed to provide an item
 *                   per every screen position.
 *
 * @param   colors   the array of colors to output. The array is supposed to provide an item per
 *                   every screen position.  This item is either -1 if this position is to be
 *                   filled with the default color or a value which is translated to the value
 *                   interpreted by the native driver.
 *
 * @param   cx       final cursor position, column.
 *
 * @param   cy       final cursor position, row.
 */
JNIEXPORT void JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_addArrayNative
  (JNIEnv *env, jobject jo, jintArray chars, jbyteArray attrs, jbyteArray colors, jint cx, jint cy)
{
   int i, j, chr, attr, color;
   jboolean isCopy = JNI_FALSE;

   jint* chrsp = (*env)->GetIntArrayElements(env, chars, &isCopy);
   jint* chrp  = chrsp;

   jint* attrsp = (*env)->GetIntArrayElements(env, attrs, &isCopy);
   jint* attrp  = attrsp;

   jint* colorsp = (*env)->GetIntArrayElements(env, colors, &isCopy);
   jint* colorp  = colorsp;

   if (chrsp == NULL || attrsp == NULL || colorsp == NULL)
   {
      // memory allocation failed for at least one of the arrays,
      // cleanup
      jint* arraysp[3] = {chrsp, attrsp, colorsp};
      jintArray arrays[3] = {chars, attrs, colors};

      int k;
      for (k = 0; k < 3; ++k)
      {
         if (arraysp[k] != NULL)
         {
            (*env)->ReleaseIntArrayElements(env, arrays[k], arraysp[k], JNI_ABORT);
         }
      }

      // and return
      return;
   }

   for (i = 0; i < base_rows; i++)
   {
      for (j = 0; j < base_cols; j++)
      {
         chr = *chrp++;
         attr = *attrp++;
         color = *colorp++;

         if (chr == -1)
         {
            continue;
         }

         attr = colorPairNative(attr, color);
         moveNative(i, j);
         addchNative(chr | attr);
       }
   }

   moveNative(cy, cx);

   syncNative();
   (*env)->ReleaseIntArrayElements(env, chars, chrsp, JNI_ABORT);
   (*env)->ReleaseIntArrayElements(env, attrs, attrsp, JNI_ABORT);
   (*env)->ReleaseIntArrayElements(env, colors, colorsp, JNI_ABORT);
}

/**
 * Add a string to the virtual terminal at the current cursor position.
 *
 * @param   env        pointer to a structure that provides access to JNI environment.
 *
 * @param   jo         reference to the Java object who called this function.
 *
 * @param   jstr       The array of characters to output.
 * 
 * @param   len        the lenght of the string to draw in characters.
 *
 * @param   attrib     the attribute to be used while drawing.
 *
 * @param   colorpair  the color value to be used while drawing.
 */
JNIEXPORT void JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_addStringNative
  (JNIEnv *env, jobject jo, jbyteArray jstr, jint len, jint attrib, jint colorpair)
{
   int i;
   
   if (len <= 0)
   {
      return;
   }

   int attr = attrib;
   attr = colorPairNative(attr, colorpair);
   
   jbyte* chrs = (jbyte*) malloc(len * sizeof(jbyte));
   
   if (chrs == NULL)
   {
      throwException(env, "java/lang/OutOfMemoryError", "Cannot copy data.");
   }
   
   (*env)->GetByteArrayRegion(env, jstr, 0, len, chrs);
   
   for (i = 0; i < len; i++)
   {
      // remove the automatic sign extension (we want to treat this as
      // an unsigned 8-bit number)
      int extended = 0x000000FF & (int) chrs[i];
      addchNative(extended | attr);
   } 

   (*env)->DeleteLocalRef(env, jstr);
   free(chrs);
}

/**
 * Beeps the tone on the terminal.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 */
JNIEXPORT void JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_beep
  (JNIEnv *env, jobject jo)
{
   beepNative();
}

/**
 * Clears the current console screen.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 */
JNIEXPORT void JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_clear
  (JNIEnv *env, jobject jo)
{
   clearNative();
}

/**
 * This method is used for getting the line-drawing special characters from underlying native
 * layer.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   class   reference to the Java class who called this function (from static context).
 *
 * @param   index   The index of the character to return. 
 *
 * @return   The character at the given index from internal array.
 */
JNIEXPORT jint JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_getACSchar
  (JNIEnv *env, jclass class, jint index)
{
   if (index >= 0 && index < ACS_MAX)
   {
      return acs_result[index];
   }
   else
   {
      return (jint) -1;
   }
}

/**
 * This method is used for getting the attributes from underlying native layer.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   class   reference to the Java class who called this function (from static context).
 *
 * @param   index   the index of the attribute to return. 
 *
 * @return   The attribute at the given index from internal array.
 */
JNIEXPORT jint JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_getAttribute
  (JNIEnv *env, jclass class, jint index)
{
   if (index >= 0 && index < ATTRIBUTE_MAX)
   {
      return attribute_result[index];
   }
   else
   {
      return (jint) -1;
   }
}

/**
 * Return the number of columns that the screen can display.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 *
 * @return   The number of columns (in characters).
 */
JNIEXPORT jint JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_getScreenColumnsNative
  (JNIEnv *env, jobject jo)
{
   return (jint) base_cols;
}

/**
 * Return the number of rows that the screen can display.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 *
 * @return   The number of rows (lines of characters).
 */
JNIEXPORT jint JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_getScreenRowsNative
  (JNIEnv *env, jobject jo)
{
   return (jint) base_rows;
}

/**
 * Gets the terminal type reported by console native backend.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   cls     reference to the Java class who called this function (from static context).
 *
 * @return   The terminal type or an empty string on any failure.
 */
JNIEXPORT jstring JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_getTermType
  (JNIEnv *env, jclass  cls)
{
   return (*env)->NewStringUTF(env, safe_termname());
}

/**
 * Get X coordinate of the cursor.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 *
 * @return   The integer value of the X coordinate.
 */
JNIEXPORT jint JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_getx
  (JNIEnv *env, jobject jo)
{
   return getx();
}

/**
 * Get Y coordinate of the cursor.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 *
 * @return   The integer value of the Y coordinate.
 */
JNIEXPORT jint JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_gety
  (JNIEnv *env, jobject jo)
{
   return gety();
}

/**
 * Check if the terminal is able to display colors or not.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 *
 * @return   <code>true</code> if display is color capable, <code>false</code> otherwise.
 */
JNIEXPORT jboolean JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_hasColors
  (JNIEnv *env, jobject jo)
{
   return hasColors();
}

/**
 * This method performs console initialization on the native side.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 */
JNIEXPORT void JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_initConsole
  (JNIEnv *env, jobject jo)
{
   char *termname;

   termname = getenv("TERM");
    
   if (termname == NULL || *termname == '\0')
   {
      termname = "vt100";
   }
   
   initConsole(env, termname);
}

/**
 * This method reads the key from keyboard by the special reading thread.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 *
 * @return   The integer code of the key that was read or -1 if there is no character to read.
 */
JNIEXPORT jint JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_readKey
  (JNIEnv *env, jobject jo)
{
   return readKey();
}

/**
 * Reset terminal settings.  May be called only if suspend() was previously called. 
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 */
JNIEXPORT void JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_resetMode
  (JNIEnv *env, jobject jo)
{
   resetMode();
}

/**
 * Re-activates the console mode after ending of the previous suspension state.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 */
JNIEXPORT void JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_resume
  (JNIEnv *env, jobject jo)
{
   resume();
}

/**
 * Set absolute cursor position
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 * 
 * @param   x       new X coordinate of the cursor.
 *
 * @param   y       new Y coordinate of the cursor.
 *
 * @return   <code>true</code> if cursor positioning succeeded.
 */
JNIEXPORT jboolean JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_setCursorNative
  (JNIEnv *env, jobject jo, jint x, jint y)
{
   return setCursorNative(x, y);
}

/**
 * Set the cursor visibility on or off. This method only affects the visibility of the cursor.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 * 
 * @param   on      when <code>true</code> the cursor will become visible, when <code>false</code>
 *                  the cursor will become invisible.
 *
 * @return   The previous state of the cursor (<code>true</code> means visible).
 */
JNIEXPORT jboolean JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_setCursorStatus
  (JNIEnv *env, jobject jo, jboolean on)
{
   return setCursorStatus(env, on);
}

/**
 * Sets the terminal type used by console backend.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   cls     reference to the Java class who called this function (from static context).
 *
 * @param   name    The new terminal type (must not be <code>null</code>).
 */
JNIEXPORT void JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_setTermType
  (JNIEnv *env, jclass  cls, jstring name)
{
   char* termname = (char*) (*env)->GetStringUTFChars(env, name, NULL);
   
   if (termname == NULL)
   {
      // out of memory!
      exit(-1);
   }
   
   restartTerminal(env, strdup(termname));
   
   (*env)->ReleaseStringUTFChars(env, name, termname);
}

/**
 * Save the current state of the console mode and switch to TTY mode temporarily.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 */
JNIEXPORT void JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_suspend
  (JNIEnv *env, jobject jo)
{
   suspend();
}

/**
 * Synchronizes the state of the console buffers with native terminal window. Meaning the kind of
 * refresh calls depending on the backend used.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 */
JNIEXPORT void JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_syncNative
   (JNIEnv *env, jobject jo)
{
   syncNative();
}
